@@ -33,17 +33,17 @@ class CORE_EXPORT QgsColorRamp
 
 #ifdef SIP_RUN
     SIP_CONVERT_TO_SUBCLASS_CODE
-    if ( sipCpp->type() == "gradient" )
+    if ( sipCpp->type() == QgsGradientColorRamp::typeString() )
       sipType = sipType_QgsGradientColorRamp;
-    else if ( sipCpp->type() == "random" )
+    else if ( sipCpp->type() == QgsLimitedRandomColorRamp::typeString() )
       sipType = sipType_QgsLimitedRandomColorRamp;
-    else if ( sipCpp->type() == "randomcolors" )
+    else if ( sipCpp->type() == QgsRandomColorRamp::typeString() )
       sipType = sipType_QgsRandomColorRamp;
-    else if ( sipCpp->type() == "preset" )
+    else if ( sipCpp->type() == QgsPresetSchemeColorRamp::typeString() )
       sipType = sipType_QgsPresetSchemeColorRamp;
-    else if ( sipCpp->type() == "colorbrewer" )
+    else if ( sipCpp->type() == QgsColorBrewerColorRamp::typeString() )
       sipType = sipType_QgsColorBrewerColorRamp;
-    else if ( sipCpp->type() == "cpt-city" )
+    else if ( sipCpp->type() == QgsCptCityColorRamp::typeString() )
       sipType = sipType_QgsCptCityColorRamp;
     else
       sipType = 0;
@@ -90,6 +90,17 @@ class CORE_EXPORT QgsColorRamp
      * Returns a string map containing all the color ramp's properties.
      */
     virtual QgsStringMap properties() const = 0;
+
+    /**
+     * Returns a list of available ramp types, where the first value in each item is the QgsColorRamp::type() string
+     * and the second is a user friendly, translated name for the color ramp type.
+     *
+     * The ramp types are returned in a order of precedence for exposing in UI, with more commonly used types
+     * listed first.
+     *
+     * \since QGIS 3.16
+     */
+    static QList< QPair< QString, QString > > rampTypes();
 };
 
 /**
@@ -144,7 +155,7 @@ class CORE_EXPORT QgsGradientColorRamp : public QgsColorRamp
      * Constructor for QgsGradientColorRamp
      * \param color1 start color, corresponding to a position of 0.0
      * \param color2 end color, corresponding to a position of 1.0
-     * \param discrete set to true for discrete interpolation instead of smoothly
+     * \param discrete set to TRUE for discrete interpolation instead of smoothly
      * interpolating between colors
      * \param stops optional list of additional color stops
      */
@@ -159,7 +170,15 @@ class CORE_EXPORT QgsGradientColorRamp : public QgsColorRamp
     int count() const override { return mStops.count() + 2; }
     double value( int index ) const override;
     QColor color( double value ) const override;
-    QString type() const override { return QStringLiteral( "gradient" ); }
+
+    /**
+     * Returns the string identifier for QgsGradientColorRamp.
+     *
+     * \since QGIS 3.16
+     */
+    static QString typeString() { return QStringLiteral( "gradient" ); }
+
+    QString type() const override;
     void invert() override;
     QgsGradientColorRamp *clone() const override SIP_FACTORY;
     QgsStringMap properties() const override;
@@ -195,7 +214,7 @@ class CORE_EXPORT QgsGradientColorRamp : public QgsColorRamp
     void setColor2( const QColor &color ) { mColor2 = color; }
 
     /**
-     * Returns true if the gradient is using discrete interpolation, rather than
+     * Returns TRUE if the gradient is using discrete interpolation, rather than
      * smoothly interpolating between colors.
      * \see setDiscrete()
      */
@@ -204,7 +223,7 @@ class CORE_EXPORT QgsGradientColorRamp : public QgsColorRamp
     /**
      * Sets whether the gradient should use discrete interpolation, rather than
      * smoothly interpolating between colors.
-     * \param discrete set to true to use discrete interpolation
+     * \param discrete set to TRUE to use discrete interpolation
      * \see convertToDiscrete()
      * \see isDiscrete()
      */
@@ -213,8 +232,8 @@ class CORE_EXPORT QgsGradientColorRamp : public QgsColorRamp
     /**
      * Converts a gradient with existing color stops to or from discrete
      * interpolation.
-     * \param discrete set to true to convert the gradient stops to discrete,
-     * or false to convert them to smooth interpolation
+     * \param discrete set to TRUE to convert the gradient stops to discrete,
+     * or FALSE to convert them to smooth interpolation
      * \see isDiscrete()
      */
     void convertToDiscrete( bool discrete );
@@ -309,7 +328,15 @@ class CORE_EXPORT QgsLimitedRandomColorRamp : public QgsColorRamp
 
     double value( int index ) const override;
     QColor color( double value ) const override;
-    QString type() const override { return QStringLiteral( "random" ); }
+
+    /**
+     * Returns the string identifier for QgsLimitedRandomColorRamp.
+     *
+     * \since QGIS 3.16
+     */
+    static QString typeString() { return QStringLiteral( "random" ); }
+
+    QString type() const override;
     QgsLimitedRandomColorRamp *clone() const override SIP_FACTORY;
     QgsStringMap properties() const override;
     int count() const override { return mCount; }
@@ -448,6 +475,13 @@ class CORE_EXPORT QgsRandomColorRamp: public QgsColorRamp
      */
     virtual void setTotalColorCount( int colorCount );
 
+    /**
+     * Returns the string identifier for QgsRandomColorRamp.
+     *
+     * \since QGIS 3.16
+     */
+    static QString typeString() { return QStringLiteral( "randomcolors" ); }
+
     QString type() const override;
 
     QgsRandomColorRamp *clone() const override SIP_FACTORY;
@@ -507,7 +541,15 @@ class CORE_EXPORT QgsPresetSchemeColorRamp : public QgsColorRamp, public QgsColo
 
     double value( int index ) const override;
     QColor color( double value ) const override;
-    QString type() const override { return QStringLiteral( "preset" ); }
+
+    /**
+     * Returns the string identifier for QgsPresetSchemeColorRamp.
+     *
+     * \since QGIS 3.16
+     */
+    static QString typeString() { return QStringLiteral( "preset" ); }
+
+    QString type() const override;
     void invert() override;
     QgsPresetSchemeColorRamp *clone() const override SIP_FACTORY;
     QgsStringMap properties() const override;
@@ -556,7 +598,15 @@ class CORE_EXPORT QgsColorBrewerColorRamp : public QgsColorRamp
 
     double value( int index ) const override;
     QColor color( double value ) const override;
-    QString type() const override { return QStringLiteral( "colorbrewer" ); }
+
+    /**
+     * Returns the string identifier for QgsColorBrewerColorRamp.
+     *
+     * \since QGIS 3.16
+     */
+    static QString typeString() { return QStringLiteral( "colorbrewer" ); }
+
+    QString type() const override { return QgsColorBrewerColorRamp::typeString(); }
     void invert() override;
     QgsColorBrewerColorRamp *clone() const override SIP_FACTORY;
     QgsStringMap properties() const override;
@@ -653,7 +703,14 @@ class CORE_EXPORT QgsCptCityColorRamp : public QgsGradientColorRamp
 
     static QgsColorRamp *create( const QgsStringMap &properties = QgsStringMap() ) SIP_FACTORY;
 
-    QString type() const override { return QStringLiteral( "cpt-city" ); }
+    /**
+     * Returns the string identifier for QgsCptCityColorRamp.
+     *
+     * \since QGIS 3.16
+     */
+    static QString typeString() { return QStringLiteral( "cpt-city" ); }
+
+    QString type() const override;
 
     void invert() override;
 
